from collections import deque


class Queue:

    def __init__(self):
        self._queue = deque()

    @property
    def length(self) -> int:
        return len(self._queue)

    def push(self, data, left: bool = False, direct_insertion: bool = False):
        try:
            if not data:
                return None
            if not direct_insertion and any(isinstance(data, t) for t in (list, tuple)):
                self._queue.extendleft(data) if left else self._queue.extend(data)
            else:
                self._queue.appendleft(data) if left else self._queue.append(data)
        except AttributeError:
            pass

    def pop(self, left: bool = True):
        try:
            return self._queue.popleft() if left else self._queue.pop()
        except IndexError:
            return None
        except AttributeError:
            return None



import time
import threading

from abc import ABC, abstractmethod
from cobweb.base import BaseItem, Queue, logger


class Pipeline(threading.Thread, ABC):

    def __init__(
            self,
            done_queue: Queue,
            upload_queue: Queue,
            upload_queue_size: int,
            upload_wait_seconds: int
    ):
        super().__init__()
        self.done_queue = done_queue
        self.upload_queue = upload_queue
        self.upload_queue_size = upload_queue_size
        self.upload_wait_seconds = upload_wait_seconds

    @abstractmethod
    def build(self, item: BaseItem) -> dict:
        pass

    @abstractmethod
    def upload(self, table: str, data: list) -> bool:
        pass

    def run(self):
        while True:
            status = self.upload_queue.length < self.upload_queue_size
            if status:
                time.sleep(self.upload_wait_seconds)
            data_info, seeds = {}, []
            for _ in range(self.upload_queue_size):
                item = self.upload_queue.pop()
                if not item:
                    break
                data = self.build(item)
                seeds.append(item.seed)
                data_info.setdefault(item.table, []).append(data)
            for table, datas in data_info.items():
                try:
                    self.upload(table, datas)
                    status = True
                except Exception as e:
                    logger.info(e)
                    status = False
                if status:
                    self.done_queue.push(seeds)





import random
import requests


class Request:

    __REQUEST_ATTRS__ = {
        "params",
        "headers",
        "cookies",
        "data",
        "json",
        "files",
        "auth",
        "timeout",
        "proxies",
        "hooks",
        "stream",
        "verify",
        "cert",
        "allow_redirects",
    }

    def __init__(
            self,
            url,
            seed,
            random_ua=True,
            check_status_code=True,
            **kwargs
    ):
        self.url = url
        self.seed = seed
        self.check_status_code = check_status_code
        self.request_setting = {}

        for k, v in kwargs.items():
            if k in self.__class__.__REQUEST_ATTRS__:
                self.request_setting[k] = v
                continue
            self.__setattr__(k, v)

        if not getattr(self, "method", None):
            self.method = "POST" if self.request_setting.get("data") or self.request_setting.get("json") else "GET"

        if random_ua:
            self._build_header()

    @property
    def _random_ua(self) -> str:
        v1 = random.randint(4, 15)
        v2 = random.randint(3, 11)
        v3 = random.randint(1, 16)
        v4 = random.randint(533, 605)
        v5 = random.randint(1000, 6000)
        v6 = random.randint(10, 80)
        user_agent = (f"Mozilla/5.0 (Macintosh; Intel Mac OS X 10_{v1}_{v2}) AppleWebKit/{v4}.{v3} "
                      f"(KHTML, like Gecko) Chrome/105.0.0.0 Safari/{v4}.{v3} Edg/105.0.{v5}.{v6}")
        return user_agent

    def _build_header(self) -> dict:
        if not self.request_setting.get("headers"):
            self.request_setting["headers"] = {"accept": "*/*", "user-agent": self._random_ua}
        elif "user-agent" not in [key.lower() for key in self.request_setting["headers"].keys()]:
            self.request_setting["headers"]["user-agent"] = self._random_ua

    def download(self) -> requests.Response:
        response = requests.request(self.method, self.url, **self.request_setting)
        if self.check_status_code:
            response.raise_for_status()
        return response




import time
import threading

from cobweb.db import RedisDB
from cobweb.base import Seed, logger
from cobweb.launchers import Launcher
from cobweb.constant import DealModel, LogTemplate


class LauncherPro(Launcher):

    def __init__(self, task, project, custom_setting=None):
        super().__init__(task, project, custom_setting)
        self._todo = "{%s:%s}:todo" % (project, task)
        self._done = "{%s:%s}:done" % (project, task)
        self._fail = "{%s:%s}:fail" % (project, task)
        self._heartbeat = "heartbeat:%s_%s" % (project, task)
        self._reset_lock = "lock:reset:%s_%s" % (project, task)
        self._heartbeat_lock = "lock:heartbeat:%s_%s" % (project, task)
        self._db = RedisDB()

        self._heartbeat_start_event = threading.Event()
        self._redis_queue_empty_event = threading.Event()

    @property
    def heartbeat(self):
        return self._db.exists(self._heartbeat)

    def _execute_heartbeat(self):
        while not self._stop.is_set():
            if self._heartbeat_start_event.is_set():
                if self._db.lock(self._heartbeat_lock, t=1):
                    self._db.setex(self._heartbeat, 3)
                time.sleep(1)
            time.sleep(0.5)

    def _reset(self):
        """
        检查过期种子，重新添加到redis缓存中
        """
        first = True
        while not self._pause.is_set():
            reset_wait_seconds = 15
            if self._db.lock(self._reset_lock, t=120):
                if not self.heartbeat:
                    self._heartbeat_start_event.set()

                _min = -int(time.time()) + self._seed_reset_seconds \
                    if self.heartbeat or not first else "-inf"

                self._db.members(
                    self._todo, 0,
                    _min=_min, _max="(0"
                )
                self._db.delete(self._reset_lock)
                reset_wait_seconds = 60

            time.sleep(reset_wait_seconds)
            first = False

    def _scheduler(self):
        """
        调度任务，获取redis队列种子，同时添加到doing字典中
        """
        if self.start_seeds:
            self.__LAUNCHER_QUEUE__['todo'].push(self.start_seeds)
        while not self._pause.is_set():
            if not self._db.zcount(self._todo, 0, "(1000"):
                time.sleep(self._scheduler_wait_seconds)
                continue
            if self.__LAUNCHER_QUEUE__['todo'].length >= self._todo_queue_size:
                time.sleep(self._todo_queue_full_wait_seconds)
                continue
            members = self._db.members(
                self._todo, int(time.time()),
                count=self._todo_queue_size,
                _min=0, _max="(1000"
            )
            for member, priority in members:
                seed = Seed(member, priority=priority)
                self.__LAUNCHER_QUEUE__['todo'].push(seed)
                self.__DOING__[seed.to_string] = seed.params.priority

    def _insert(self):
        """
        添加新种子到redis队列中
        """
        while not self._pause.is_set():
            seeds = {}
            status = self.__LAUNCHER_QUEUE__['new'].length < self._new_queue_max_size
            for _ in range(self._new_queue_max_size):
                seed = self.__LAUNCHER_QUEUE__['new'].pop()
                if not seed:
                    break
                seeds[seed.to_string] = seed.params.priority
            if seeds:
                self._db.zadd(self._todo, seeds, nx=True)
            if status:
                time.sleep(self._new_queue_wait_seconds)

    def _refresh(self):
        """
        刷新doing种子过期时间，防止reset重新消费
        """
        while not self._pause.is_set():
            if self.__DOING__:
                refresh_time = int(time.time())
                seeds = {k:-refresh_time - v / 1000 for k, v in self.__DOING__.items()}
                self._db.zadd(self._todo, item=seeds, xx=True)
            time.sleep(30)

    def _delete(self):
        """
        删除队列种子，根据状态添加至成功或失败队列，移除doing字典种子索引
        """
        while not self._pause.is_set():
            seeds, s_seeds, f_seeds = [], [], []
            status = self.__LAUNCHER_QUEUE__['done'].length < self._done_queue_max_size

            for _ in range(self._done_queue_max_size):
                seed = self.__LAUNCHER_QUEUE__['done'].pop()
                if not seed:
                    break
                if seed.params.identifier == DealModel.fail:
                    f_seeds.append(seed.to_string)
                elif self._done_model == 1:
                    s_seeds.append(seed.to_string)
                else:
                    seeds.append(seed.to_string)
            if seeds:
                self._db.zrem(self._todo, *seeds)
            if s_seeds:
                self._db.done([self._todo, self._done], *s_seeds)
            if f_seeds:
                self._db.done([self._todo, self._fail], *f_seeds)

            self._remove_doing_seeds(seeds)

            if status:
                time.sleep(self._done_queue_wait_seconds)

    def _polling(self):
        check_emtpy_times = 0
        while not self._stop.is_set():
            queue_not_empty_count = 0
            pooling_wait_seconds = 30
            if not self._db.zcard(self._todo):
                for q in self.__LAUNCHER_QUEUE__.values():
                    if q.length != 0:
                        queue_not_empty_count += 1
                if self._pause.is_set() and queue_not_empty_count != 0:
                    self._pause.clear()
                    self._execute()
                elif queue_not_empty_count == 0:
                    pooling_wait_seconds = 3
                    check_emtpy_times += 1
                else:
                    check_emtpy_times = 0
                if check_emtpy_times > 2:
                    check_emtpy_times = 0
                    self.__DOING__ = {}
                    self._pause.set()
            if not self._pause.is_set():
                logger.info(LogTemplate.launcher_pro_polling.format(
                    task=self.task,
                    doing_len=len(self.__DOING__.keys()),
                    todo_len=self.__LAUNCHER_QUEUE__['todo'].length,
                    done_len=self.__LAUNCHER_QUEUE__['done'].length,
                    redis_seed_count=self._db.zcount(self._todo, "-inf", "+inf"),
                    redis_todo_len=self._db.zcount(self._todo, 0, "(1000"),
                    redis_doing_len=self._db.zcount(self._todo, "-inf", "(0"),
                    upload_len=self._upload_queue.length
                ))
            time.sleep(pooling_wait_seconds)



import os

# redis db config
REDIS_CONFIG = {
    "host": os.getenv("REDIS_HOST"),
    "password": os.getenv("REDIS_PASSWORD"),
    "port": int(os.getenv("REDIS_PORT", 6379)),
    "db": int(os.getenv("REDIS_DB", 0)),
}

# loghub db config
LOGHUB_TOPIC = os.getenv("LOGHUB_TOPIC")
LOGHUB_SOURCE = os.getenv("LOGHUB_SOURCE")
LOGHUB_PROJECT = os.getenv("LOGHUB_PROJECT")
LOGHUB_CONFIG = {
    "endpoint": os.getenv("LOGHUB_ENDPOINT"),
    "accessKeyId": os.getenv("LOGHUB_ACCESS_KEY"),
    "accessKey": os.getenv("LOGHUB_SECRET_KEY")
}

# oss util config
OSS_BUCKET = os.getenv("OSS_BUCKET")
OSS_ENDPOINT = os.getenv("OSS_ENDPOINT")
OSS_ACCESS_KEY = os.getenv("OSS_ACCESS_KEY")
OSS_SECRET_KEY = os.getenv("OSS_SECRET_KEY")
OSS_MIN_UPLOAD_SIZE = 1024
OSS_CHUNK_SIZE = 1024 ** 2

# 采集器选择
CRAWLER = "cobweb.crawlers.CrawlerAir"

# 数据上传链路
PIPELINE = "cobweb.pipelines.loghub_pipeline.LoghubPipeline"


# Launcher 等待时间
SCHEDULER_WAIT_SECONDS = 15  # 调度等待时间
TODO_QUEUE_FULL_WAIT_SECONDS = 5  # todo队列已满时等待时间
NEW_QUEUE_WAIT_SECONDS = 30   # new队列等待时间
DONE_QUEUE_WAIT_SECONDS = 15   # done队列等待时间
UPLOAD_QUEUE_WAIT_SECONDS = 15   # upload队列等待时间
SEED_RESET_SECONDS = 600   # 种子重制时间


# Launcher 队列长度
TODO_QUEUE_SIZE = 100  # todo队列长度
NEW_QUEUE_MAX_SIZE = 100  # new队列长度
DONE_QUEUE_MAX_SIZE = 100  # done队列长度
UPLOAD_QUEUE_MAX_SIZE = 100  # upload队列长度

# DONE_MODEL IN (0, 1), 种子完成模式
DONE_MODEL = 0   # 0:种子消费成功直接从队列移除，失败则添加至失败队列；1:种子消费成功添加至成功队列，失败添加至失败队列

# DOWNLOAD_MODEL IN (0, 1), 下载模式
DOWNLOAD_MODEL = 0  # 0: 通用下载；1:文件下载

# spider
SPIDER_THREAD_NUM = 10
SPIDER_MAX_RETRIES = 5

# 文件下载响应类型过滤
FILE_FILTER_CONTENT_TYPE = ["text/html", "application/xhtml+xml"]




class Response:

    def __init__(
            self,
            seed,
            response,
            **kwargs
    ):
        self.seed = seed
        self.response = response

        for k, v in kwargs.items():
            self.__setattr__(k, v)

    @property
    def to_dict(self):
        _dict = self.__dict__.copy()
        _dict.pop('seed')
        _dict.pop('response')
        return _dict




from typing import Union
from cobweb import setting
from cobweb.utils import OssUtil
from cobweb.crawlers import Crawler
from cobweb.base import Seed, BaseItem, Request, Response
from cobweb.exceptions import OssDBPutPartError, OssDBMergeError


class CrawlerAir(Crawler):
    
    oss_util = OssUtil()

    @staticmethod
    def download(item: Request) -> Union[Seed, BaseItem, Response, str]:
        seed_dict = item.seed.to_dict
        bucket_name = CrawlerAir.oss_util.bucket
        try:
            key = item.seed.oss_path
            if CrawlerAir.oss_util.exists(key):
                content_length = CrawlerAir.oss_util.head(key).content_length
                yield Response(item.seed, "exists", bucket_name=bucket_name, data_size=content_length, **seed_dict)

            end = seed_dict.get("end", "")
            start = seed_dict.get("start", "0")

            if end or int(start):
                item.request_setting["headers"]['Range'] = f'bytes={start}-{end}'

            if not item.seed.params.identifier:
                content = b""
                chunk_size = CrawlerAir.oss_util.chunk_size
                min_upload_size = CrawlerAir.oss_util.min_upload_size
                position = seed_dict.get("position", 1)

                response = item.download()

                content_length = response.headers.get("content-length") or 0
                content_type = response.headers.get("content-type", "").split(";")[0]
                if content_type and content_type in setting.FILE_FILTER_CONTENT_TYPE:
                    yield Response(
                        item.seed, response, filter=True, msg=f"response content type is {content_type}",
                        bucket_name=bucket_name, data_size=content_length, **seed_dict
                    )
                elif position == 1 and min_upload_size >= int(content_length) > 0:
                    """过小文件标识返回"""
                    yield Response(
                        item.seed, response, filter=True, msg="file size is too small",
                        bucket_name=bucket_name, data_size=content_length, **seed_dict
                    )
                elif position == 1 and chunk_size > int(content_length) > min_upload_size:
                    """小文件直接下载"""
                    for part_data in response.iter_content(chunk_size):
                        content += part_data
                    CrawlerAir.oss_util.put(key, content)
                    yield Response(item.seed, response, bucket_name=bucket_name, data_size=content_length, **seed_dict)
                    response.close()
                else:
                    """中大文件同步分片下载"""
                    upload_content_length = 0
                    if not seed_dict.get("upload_id"):
                        seed_dict["upload_id"] = CrawlerAir.oss_util.init_part(key).upload_id
                    upload_id = seed_dict["upload_id"]
                    for part_data in response.iter_content(chunk_size):
                        content += part_data
                        if len(content) >= chunk_size:
                            upload_data = content[:chunk_size]
                            content = content[chunk_size:]
                            CrawlerAir.oss_util.put_part(key, upload_id, position, upload_data)
                            upload_content_length += len(upload_data)
                            position += 1
                            seed_dict['position'] = position
                            seed_dict['start'] = upload_content_length

                    response.close()
                    if content:
                        CrawlerAir.oss_util.put_part(key, upload_id, position, content)
                        content_length += len(content)
                    CrawlerAir.oss_util.merge(key, upload_id)
                    yield Response(item.seed, response, bucket_name=bucket_name, data_size=content_length, **seed_dict)

            elif item.seed.params.identifier == "merge":
                CrawlerAir.oss_util.merge(key, seed_dict["upload_id"])
                content_length = CrawlerAir.oss_util.head(key).content_length
                yield Response(item.seed, "merge", bucket_name=bucket_name, data_size=content_length, **seed_dict)
        except OssDBPutPartError:
            yield Seed(seed_dict)
        except OssDBMergeError:
            yield Seed(seed_dict, identifier="merge")


class CrawlerPro(Crawler):

    oss_util = OssUtil()

    @staticmethod
    def download(item: Request) -> Union[Seed, BaseItem, Response, str]:
        seed_dict = item.seed.to_dict
        bucket_name = CrawlerAir.oss_util.bucket
        try:
            key = item.seed.oss_path
            if CrawlerAir.oss_util.exists(key):
                content_length = CrawlerAir.oss_util.head(key).content_length
                yield Response(item.seed, "exists", bucket_name=bucket_name, data_size=content_length, **seed_dict)

            end = seed_dict.get("end", "")
            start = seed_dict.get("start", "0")

            if end or int(start):
                item.request_setting["headers"]['Range'] = f'bytes={start}-{end}'

            if not item.seed.params.identifier:
                content = b""
                chunk_size = CrawlerAir.oss_util.chunk_size
                min_upload_size = CrawlerAir.oss_util.min_upload_size
                position = seed_dict.get("position", 1)

                response = item.download()

                content_length = response.headers.get("content-length") or 0
                content_type = response.headers.get("content-type", "").split(";")[0]
                if content_type and content_type in setting.FILE_FILTER_CONTENT_TYPE:
                    yield Response(
                        item.seed, response, filter=True, msg=f"response content type is {content_type}",
                        bucket_name=bucket_name, data_size=content_length, **seed_dict
                    )
                elif position == 1 and min_upload_size >= int(content_length) > 0:
                    """过小文件标识返回"""
                    yield Response(
                        item.seed, response, filter=True, msg="file size is too small",
                        bucket_name=bucket_name, data_size=content_length, **seed_dict
                    )
                elif position == 1 and chunk_size > int(content_length) > min_upload_size:
                    """小文件直接下载"""
                    for part_data in response.iter_content(chunk_size):
                        content += part_data
                    CrawlerAir.oss_util.put(key, content)
                    yield Response(item.seed, response, bucket_name=bucket_name, data_size=content_length, **seed_dict)
                    response.close()
                else:
                    """中大文件同步分片下载"""
                    upload_content_length = 0
                    if not seed_dict.get("upload_id"):
                        seed_dict["upload_id"] = CrawlerAir.oss_util.init_part(key).upload_id
                    upload_id = seed_dict["upload_id"]
                    for part_data in response.iter_content(chunk_size):
                        content += part_data
                        if len(content) >= chunk_size:
                            upload_data = content[:chunk_size]
                            content = content[chunk_size:]
                            CrawlerAir.oss_util.put_part(key, upload_id, position, upload_data)
                            upload_content_length += len(upload_data)
                            position += 1
                            seed_dict['position'] = position
                            seed_dict['start'] = upload_content_length

                    response.close()
                    if content:
                        CrawlerAir.oss_util.put_part(key, upload_id, position, content)
                        content_length += len(content)
                    CrawlerAir.oss_util.merge(key, upload_id)
                    yield Response(item.seed, response, bucket_name=bucket_name, data_size=content_length, **seed_dict)
                    # data, cols = download_meta(item.seed, bucket_name, data_size=content_length, **seed_dict)
                    # yield DownloadItem(item.seed, sid=item.seed.sid, cols=cols, data=data)

            elif item.seed.params.identifier == "merge":
                CrawlerAir.oss_util.merge(key, seed_dict["upload_id"])
                content_length = CrawlerAir.oss_util.head(key).content_length
                yield Response(item.seed, "merge", bucket_name=bucket_name, data_size=content_length, **seed_dict)
                # data, cols = download_meta(item.seed, bucket_name, data_size=content_length, **seed_dict)
                # yield DownloadItem(item.seed, sid=item.seed.sid, cols=cols, data=data)
        except OssDBPutPartError:
            yield Seed(seed_dict)
        except OssDBMergeError:
            yield Seed(seed_dict, identifier="merge")



import time
import inspect
import threading
import importlib

from cobweb import setting
from cobweb.base import Seed, Queue
from cobweb.utils.tools import dynamic_load_class


class Launcher(threading.Thread):

    SEEDS = []

    __DOING__ = {}

    __CUSTOM_FUNC__ = {
        "download": None,
        "download_midware": None,
        "parse": None,
    }

    __LAUNCHER_QUEUE__ = {
        "new": Queue(),
        "todo": Queue(),
        "done": Queue(),
    }

    __LAUNCHER_FUNC__ = [
        "_reset",
        "_scheduler",
        "_insert",
        "_refresh",
        "_delete",
    ]

    def __init__(self, task, project, custom_setting=None):
        super().__init__()
        self.task = task
        self.project = project

        self._stop = threading.Event()  # 结束事件
        self._pause = threading.Event()  # 暂停事件

        if custom_setting:
            setting_ = dict()
            if isinstance(custom_setting, dict):
                setting_ = custom_setting
            else:
                if isinstance(custom_setting, str):
                    custom_setting = importlib.import_module(custom_setting)
                if not inspect.ismodule(custom_setting):
                    raise Exception
                for k, v in custom_setting.__dict__.items():
                    if not k.startswith("__") and not inspect.ismodule(v):
                        setting_[k] = v
            for k, v in setting_.items():
                setattr(setting, k, v)

        self._Crawler = dynamic_load_class(setting.CRAWLER)
        self._Pipeline = dynamic_load_class(setting.PIPELINE)

        self._scheduler_wait_seconds = setting.SCHEDULER_WAIT_SECONDS
        self._todo_queue_full_wait_seconds = setting.TODO_QUEUE_FULL_WAIT_SECONDS
        self._new_queue_wait_seconds = setting.NEW_QUEUE_WAIT_SECONDS
        self._done_queue_wait_seconds = setting.DONE_QUEUE_WAIT_SECONDS
        self._upload_queue_wait_seconds = setting.UPLOAD_QUEUE_WAIT_SECONDS
        self._seed_reset_seconds = setting.SEED_RESET_SECONDS

        self._todo_queue_size = setting.TODO_QUEUE_SIZE
        self._new_queue_max_size = setting.NEW_QUEUE_MAX_SIZE
        self._done_queue_max_size = setting.DONE_QUEUE_MAX_SIZE
        self._upload_queue_max_size = setting.UPLOAD_QUEUE_MAX_SIZE

        self._done_model = setting.DONE_MODEL

        self._upload_queue = Queue()

    @property
    def start_seeds(self):
        return [Seed(seed) for seed in self.SEEDS]

    @property
    def request(self):
        """
        自定义request函数
        use case:
            from cobweb.base import Request, BaseItem
            @launcher.request
            def request(seed: Seed) -> Union[Request, BaseItem]:
                ...
                return Request(seed.url, seed)
        """
        def decorator(func):
            self.__CUSTOM_FUNC__["request"] = func
        return decorator

    @property
    def download(self):
        """
        自定义download函数
        use case:
            from cobweb.base import Request, Response, Seed, BaseItem
            @launcher.download
            def download(item: Request) -> Union[Seed, BaseItem, Response, str]:
                ...
                yield Response(item.seed, response)
        """
        def decorator(func):
            self.__CUSTOM_FUNC__["download"] = func
        return decorator

    @property
    def parse(self):
        """
        自定义parse函数, xxxItem为自定义的存储数据类型
        use case:
            from cobweb.base import Request, Response
            @launcher.download
            def download(item: Response) -> BaseItem:
               ...
               yield xxxItem(seed, **kwargs)
        """
        def decorator(func):
            self.__CUSTOM_FUNC__["parse"] = func
        return decorator

    def _remove_doing_seeds(self, seeds):
        for seed in seeds:
            self.__DOING__.pop(seed, None)

    def _execute_heartbeat(self):
        pass

    def _reset(self):
        pass

    def _scheduler(self):
        pass

    def _insert(self):
        pass

    def _refresh(self):
        pass

    def _delete(self):
        pass

    def _execute(self):
        for func_name in self.__LAUNCHER_FUNC__:
            threading.Thread(name=func_name, target=getattr(self, func_name)).start()
            time.sleep(2)

    def _polling(self):

        check_emtpy_times = 0

        while not self._stop.is_set():

            queue_not_empty_count = 0

            for q in self.__LAUNCHER_QUEUE__.values():
                if q.length != 0:
                    queue_not_empty_count += 1

            if self._pause.is_set() and queue_not_empty_count != 0:
                self._pause.clear()
                self._execute()

            elif queue_not_empty_count == 0:
                check_emtpy_times += 1
            else:
                check_emtpy_times = 0

            if check_emtpy_times > 2:
                check_emtpy_times = 0
                self.__DOING__ = {}
                self._pause.set()

    def run(self):
        threading.Thread(target=self._execute_heartbeat).start()

        self._Crawler(
            upload_queue=self._upload_queue,
            custom_func=self.__CUSTOM_FUNC__,
            launcher_queue=self.__LAUNCHER_QUEUE__,
        ).start()

        self._Pipeline(
            upload_queue=self._upload_queue,
            done_queue=self.__LAUNCHER_QUEUE__["done"],
            upload_queue_size=self._upload_queue_max_size,
            upload_wait_seconds=self._upload_queue_wait_seconds
        ).start()

        self._execute()
        self._polling()



from .common_queue import Queue
from .response import Response
from .request import Request
from .item import BaseItem
from .seed import Seed

from .log import logger
from .decorators import decorator_oss_db






class LauncherModel:
    task = "launcher model: task"
    resident = "launcher model: resident"


class DownloadModel:
    common = "download model: common"
    file = "download model: file"


class LogModel:
    simple = "log model: simple"
    common = "log model: common"
    detailed = "log model: detailed"


class DealModel:
    fail = "deal model: fail"
    done = "deal model: done"
    poll = "deal model: poll"


class LogTemplate:

    launcher_pro_polling = """
----------------------- start - 轮训日志: {task} -----------------
        内存队列 
            种子数:  {doing_len}
            待消费:  {todo_len}
            已消费:  {done_len}
        redis队列 
            种子数:  {redis_seed_count}
            待消费:  {redis_todo_len}
            消费中:  {redis_doing_len}
        存储队列 
            待上传:  {upload_len}
----------------------- end  - 轮训日志: {task} ------------------
"""

    download_exception = """
----------------------- download exception -----------------------
        种子详情 \n{detail}
        种子参数
            retry         :    {retry}
            priority      :    {priority}
            seed_version  :    {seed_version}
            identifier    :    {identifier}
        exception
            msg           :    {exception}
------------------------------------------------------------------
"""

    download_info = """
------------------------ download info ---------------------------
        种子详情 \n{detail}
        种子参数
            retry         :    {retry}
            priority      :    {priority}
            seed_version  :    {seed_version}
            identifier    :    {identifier}
        response 
            status        :    {status} \n{response}
------------------------------------------------------------------
"""


import threading

from inspect import isgenerator
from typing import Union, Callable, Mapping

from cobweb.base import Queue, Seed, BaseItem, Request, Response, logger
from cobweb.constant import DealModel, LogTemplate
from cobweb.utils import download_log_info
from cobweb import setting


class Crawler(threading.Thread):

    def __init__(
            self,
            upload_queue: Queue,
            custom_func: Union[Mapping[str, Callable]],
            launcher_queue: Union[Mapping[str, Queue]],
    ):
        super().__init__()

        self.upload_queue = upload_queue
        for func_name, _callable in custom_func.items():
            if isinstance(_callable, Callable):
                self.__setattr__(func_name, _callable)

        self.launcher_queue = launcher_queue

        self.spider_thread_num = setting.SPIDER_THREAD_NUM
        self.max_retries = setting.SPIDER_MAX_RETRIES

    @staticmethod
    def request(seed: Seed) -> Union[Request, BaseItem]:
        stream = True if setting.DOWNLOAD_MODEL else False
        return Request(seed.url, seed, stream=stream, timeout=5)

    @staticmethod
    def download(item: Request) -> Union[Seed, BaseItem, Response, str]:
        response = item.download()
        yield Response(item.seed, response)

    @staticmethod
    def parse(item: Response) -> BaseItem:
        pass

    def get(self) -> Seed:
        return self.launcher_queue['todo'].pop()

    def spider(self):
        while True:
            seed = self.get()

            if not seed:
                continue

            elif seed.params.retry >= self.max_retries:
                self.launcher_queue['done'].push(seed)
                continue

            item = self.request(seed)

            if isinstance(item, Request):

                download_iterators = self.download(item)

                if not isgenerator(download_iterators):
                    raise TypeError("download function isn't a generator")

                seed_detail_log_info = download_log_info(seed.to_dict)

                try:
                    for it in download_iterators:
                        if isinstance(it, Response):
                            response_detail_log_info = download_log_info(it.to_dict)
                            logger.info(LogTemplate.download_info.format(
                                detail=seed_detail_log_info, retry=item.seed.params.retry,
                                priority=item.seed.params.priority,
                                seed_version=item.seed.params.seed_version,
                                identifier=item.seed.params.identifier,
                                status=it.response, response=response_detail_log_info
                            ))
                            parse_iterators = self.parse(it)
                            if not isgenerator(parse_iterators):
                                raise TypeError("parse function isn't a generator")
                            for upload_item in parse_iterators:
                                if not isinstance(upload_item, BaseItem):
                                    raise TypeError("upload_item isn't BaseItem subclass")
                                self.upload_queue.push(upload_item)
                        elif isinstance(it, BaseItem):
                            self.upload_queue.push(it)
                        elif isinstance(it, Seed):
                            self.launcher_queue['new'].push(it)
                        elif isinstance(it, str) and it == DealModel.poll:
                            self.launcher_queue['todo'].push(item)
                            break
                        elif isinstance(it, str) and it == DealModel.done:
                            self.launcher_queue['done'].push(seed)
                            break
                        elif isinstance(it, str) and it == DealModel.fail:
                            seed.params.identifier = DealModel.fail
                            self.launcher_queue['done'].push(seed)
                            break
                        else:
                            raise TypeError("yield value type error!")

                except Exception as e:
                    logger.info(LogTemplate.download_exception.format(
                        detail=seed_detail_log_info, retry=seed.params.retry,
                        priority=seed.params.priority, seed_version=seed.params.seed_version,
                        identifier=seed.params.identifier, exception=e
                    ))
                    seed.params.retry += 1
                    self.launcher_queue['todo'].push(seed)

            elif isinstance(item, BaseItem):
                self.upload_queue.push(item)

    def run(self):
        for index in range(self.spider_thread_num):
            threading.Thread(name=f"spider_{index}", target=self.spider).start()

